from dbt.logger import GLOBAL_LOGGER as logger, log_cache_events, log_manager

import argparse
import os.path
import sys
import traceback
from contextlib import contextmanager

import dbt.version
from dbt.profiler import profiler
from dbt.adapters.factory import reset_adapters, cleanup_connections

import dbt.tracking

from dbt.utils import ExitCodes
from dbt.config.profile import read_user_config
from dbt.exceptions import (
    RuntimeException,
    InternalException,
    NotImplementedException,
    FailedToConnectException
)
import dbt.flags as flags

from dbt_rpc.task.server import RPCServerTask


def initialize_tracking_from_flags():
    # NOTE: this is copied from dbt-core
    # Setting these used to be in UserConfig, but had to be moved here
    if flags.SEND_ANONYMOUS_USAGE_STATS:
        dbt.tracking.initialize_tracking(flags.PROFILES_DIR)
    else:
        dbt.tracking.do_not_track()


class DBTVersion(argparse.Action):
    """This is very very similar to the builtin argparse._Version action,
    except it just calls dbt.version.get_version_information().
    """

    def __init__(self,
                 option_strings,
                 version=None,
                 dest=argparse.SUPPRESS,
                 default=argparse.SUPPRESS,
                 help="show program's version number and exit"):
        super().__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help)

    def __call__(self, parser, namespace, values, option_string=None):
        formatter = argparse.RawTextHelpFormatter(prog=parser.prog)
        formatter.add_text(dbt.version.get_version_information())
        parser.exit(message=formatter.format_help())


class DBTArgumentParser(argparse.ArgumentParser):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.register('action', 'dbtversion', DBTVersion)

    def add_optional_argument_inverse(
        self,
        name,
        *,
        enable_help=None,
        disable_help=None,
        dest=None,
        no_name=None,
        default=None,
    ):
        mutex_group = self.add_mutually_exclusive_group()
        if not name.startswith('--'):
            raise InternalException(
                'cannot handle optional argument without "--" prefix: '
                f'got "{name}"'
            )
        if dest is None:
            dest_name = name[2:].replace('-', '_')
        else:
            dest_name = dest

        if no_name is None:
            no_name = f'--no-{name[2:]}'

        mutex_group.add_argument(
            name,
            action='store_const',
            const=True,
            dest=dest_name,
            default=default,
            help=enable_help,
        )

        mutex_group.add_argument(
            f'--no-{name[2:]}',
            action='store_const',
            const=False,
            dest=dest_name,
            default=default,
            help=disable_help,
        )

        return mutex_group


class RPCArgumentParser(DBTArgumentParser):
    def exit(self, status=0, message=None):
        if status == 0:
            return
        else:
            raise TypeError(message)


def main(args=None):
    if args is None:
        args = sys.argv[1:]
    with log_manager.applicationbound():
        try:
            results, succeeded = handle_and_check(args)
            if succeeded:
                exit_code = ExitCodes.Success.value
            else:
                exit_code = ExitCodes.ModelError.value

        except KeyboardInterrupt:
            logger.info("ctrl-c")
            exit_code = ExitCodes.UnhandledError.value

        # This can be thrown by eg. argparse
        except SystemExit as e:
            exit_code = e.code

        except BaseException as e:
            logger.warning("Encountered an error:")
            logger.warning(str(e))

            if log_manager.initialized:
                logger.debug(traceback.format_exc())
            elif not isinstance(e, RuntimeException):
                # if it did not come from dbt proper and the logger is not
                # initialized (so there's no safe path to log to), log the
                # stack trace at error level.
                logger.error(traceback.format_exc())
            exit_code = ExitCodes.UnhandledError.value

    sys.exit(exit_code)


# here for backwards compatibility
def handle(args):
    res, success = handle_and_check(args)
    return res


@contextmanager
def adapter_management():
    reset_adapters()
    try:
        yield
    finally:
        cleanup_connections()


def handle_and_check(args):
    with log_manager.applicationbound():
        parsed = parse_args(args)

        # Set flags from args, user config, and env vars
        user_config = read_user_config(flags.PROFILES_DIR)  # This is read again later
        flags.set_from_args(parsed, user_config)
        initialize_tracking_from_flags()
        # Set log_format from flags
        parsed.cls.set_log_format()

        # we've parsed the args and set the flags - we can now decide if we're debug or not
        if flags.DEBUG:
            log_manager.set_debug()

        profiler_enabled = False

        if parsed.record_timing_info:
            profiler_enabled = True

        with profiler(
            enable=profiler_enabled,
            outfile=parsed.record_timing_info
        ):

            with adapter_management():

                task, res = run_from_args(parsed)
                success = task.interpret_results(res)

            return res, success


@contextmanager
def track_run(task):
    dbt.tracking.track_invocation_start(config=task.config, args=task.args)
    try:
        yield
        dbt.tracking.track_invocation_end(
            config=task.config, args=task.args, result_type="ok"
        )
    except (NotImplementedException,
            FailedToConnectException) as e:
        logger.error('ERROR: {}'.format(e))
        dbt.tracking.track_invocation_end(
            config=task.config, args=task.args, result_type="error"
        )
    except Exception:
        dbt.tracking.track_invocation_end(
            config=task.config, args=task.args, result_type="error"
        )
        raise
    finally:
        dbt.tracking.flush()


def run_from_args(parsed):
    log_cache_events(getattr(parsed, 'log_cache_events', False))

    # we can now use the logger for stdout
    # set log_format in the logger
    parsed.cls.pre_init_hook(parsed)

    logger.info("Running with dbt{}".format(dbt.version.installed))

    # this will convert DbtConfigErrors into RuntimeExceptions
    # task could be any one of the task objects
    task = parsed.cls.from_args(args=parsed)

    logger.debug("running dbt with arguments {parsed}", parsed=str(parsed))

    log_path = None
    if task.config is not None:
        log_path = getattr(task.config, 'log_path', None)
    # we can finally set the file logger up
    log_manager.set_path(log_path)
    if dbt.tracking.active_user is not None:  # mypy appeasement, always true
        logger.debug("Tracking: {}".format(dbt.tracking.active_user.state()))

    results = None

    with track_run(task):
        results = task.run()

    return task, results


def _build_base_subparser():
    base_subparser = argparse.ArgumentParser(add_help=False)

    base_subparser.add_argument(
        '--project-dir',
        default=None,
        type=str,
        help='''
        Which directory to look in for the dbt_project.yml file.
        Default is the current working directory and its parents.
        '''
    )

    base_subparser.add_argument(
        '--profiles-dir',
        default=None,
        dest='sub_profiles_dir',  # Main cli arg precedes subcommand
        type=str,
        help='''
        Which directory to look in for the profiles.yml file.
        If not set, dbt will look in the current working directory first, then HOME/.dbt/
        '''
    )

    base_subparser.add_argument(
        '--profile',
        required=False,
        type=str,
        help='''
        Which profile to load. Overrides setting in dbt_project.yml.
        '''
    )

    base_subparser.add_argument(
        '-t',
        '--target',
        default=None,
        type=str,
        help='''
        Which target to load for the given profile
        ''',
    )

    base_subparser.add_argument(
        '--vars',
        type=str,
        default='{}',
        help='''
        Supply variables to the project. This argument overrides variables
        defined in your dbt_project.yml file. This argument should be a YAML
        string, eg. '{my_variable: my_value}'
        '''
    )

    # if set, log all cache events. This is extremely verbose!
    base_subparser.add_argument(
        '--log-cache-events',
        action='store_true',
        help=argparse.SUPPRESS,
    )

    base_subparser.set_defaults(defer=None, state=None)
    return base_subparser


def parse_args(args, cls=DBTArgumentParser):
    p = cls(
        prog='dbt-rpc',
        description='''
        An ELT tool for managing your SQL transformations and data models.
        For more documentation on these commands, visit: docs.getdbt.com
        ''',
        epilog='''
        Specify one of these sub-commands and you can find more help from
        there.
        '''
    )

    p.add_argument(
        '--version',
        action='dbtversion',
        help='''
        Show version information
        ''')

    p.add_argument(
        '-r',
        '--record-timing-info',
        default=None,
        type=str,
        help='''
        When this option is passed, dbt will output low-level timing stats to
        the specified file. Example: `--record-timing-info output.profile`
        '''
    )

    p.add_argument(
        '-d',
        '--debug',
        action='store_true',
        default=None,
        help='''
        Display debug logging during dbt execution. Useful for debugging and
        making bug reports.
        '''
    )

    p.add_argument(
        '--log-format',
        choices=['text', 'json', 'default'],
        default=None,
        help='''Specify the log format, overriding the command's default.'''
    )

    p.add_argument(
        '--no-write-json',
        action='store_false',
        default=None,
        dest='write_json',
        help='''
        If set, skip writing the manifest and run_results.json files to disk
        '''
    )
    colors_flag = p.add_mutually_exclusive_group()
    colors_flag.add_argument(
        '--use-colors',
        action='store_const',
        const=True,
        default=None,
        dest='use_colors',
        help='''
        Colorize the output DBT prints to the terminal. Output is colorized by
        default and may also be set in a profile or at the command line.
        Mutually exclusive with --no-use-colors
        '''
    )
    colors_flag.add_argument(
        '--no-use-colors',
        action='store_const',
        const=False,
        dest='use_colors',
        help='''
        Do not colorize the output DBT prints to the terminal. Output is
        colorized by default and may also be set in a profile or at the
        command line.
        Mutually exclusive with --use-colors
        '''
    )

    p.add_argument(
        '--printer-width',
        dest='printer_width',
        help='''
        Sets the width of terminal output
        '''
    )

    p.add_argument(
        '--warn-error',
        action='store_true',
        default=None,
        help='''
        If dbt would normally warn, instead raise an exception. Examples
        include --models that selects nothing, deprecations, configurations
        with no associated models, invalid test configurations, and missing
        sources/refs in tests.
        '''
    )

    p.add_argument(
        '--no-version-check',
        dest='version_check',
        action='store_false',
        default=None,
        help='''
        If set, skip ensuring dbt's version matches the one specified in
        the dbt_project.yml file ('require-dbt-version')
        '''
    )

    p.add_optional_argument_inverse(
        '--partial-parse',
        enable_help='''
        Allow for partial parsing by looking for and writing to a pickle file
        in the target directory. This overrides the user configuration file.
        ''',
        disable_help='''
        Disallow partial parsing. This overrides the user configuration file.
        ''',
    )

    # if set, run dbt in single-threaded mode: thread count is ignored, and
    # calls go through `map` instead of the thread pool. This is useful for
    # getting performance information about aspects of dbt that normally run in
    # a thread, as the profiler ignores child threads. Users should really
    # never use this.
    p.add_argument(
        '--single-threaded',
        action='store_true',
        help=argparse.SUPPRESS,
    )

    # if set, will use the latest features from the static parser instead of
    # the stable static parser.
    p.add_argument(
        '--use-experimental-parser',
        action='store_true',
        default=None,
        help='''
        Enables experimental parsing features.
        '''
    )

    # if set, will disable the use of the stable static parser and instead
    # always rely on jinja rendering.
    p.add_argument(
        '--no-static-parser',
        default=None,
        dest='static_parser',
        action='store_false',
        help='''
        Disables the static parser.
        '''
    )

    p.add_argument(
        '--profiles-dir',
        default=None,
        dest='profiles_dir',
        type=str,
        help='''
        Which directory to look in for the profiles.yml file.
        If not set, dbt will look in the current working directory first, then HOME/.dbt/
        '''
    )

    p.add_argument(
        '--no-anonymous-usage-stats',
        action='store_false',
        default=None,
        dest='send_anonymous_usage_stats',
        help='''
        Do not send anonymous usage stat to dbt Labs
        '''
    )

    p.add_argument(
        '-x',
        '--fail-fast',
        dest='fail_fast',
        action='store_true',
        default=None,
        help='''
        Stop execution upon a first failure.
        '''
    )

    subs = p.add_subparsers(title="Available sub-commands")

    base_subparser = _build_base_subparser()

    rpc_sub = subs.add_parser(
        'serve',
        parents=[base_subparser],
        help='''
        Start a json-rpc server
        ''',
    )
    rpc_sub.add_argument(
        '--host',
        default='0.0.0.0',
        help='''
        Specify the host to listen on for the rpc server.
        ''',
    )
    rpc_sub.add_argument(
        '--port',
        default=8580,
        type=int,
        help='''
        Specify the port number for the rpc server.
        ''',
    )
    rpc_sub.set_defaults(cls=RPCServerTask, which='rpc', rpc_method=None)
    # the rpc task does a 'compile', so we need these attributes to exist, but
    # we don't want users to be allowed to set them.
    rpc_sub.set_defaults(models=None, exclude=None)

    rpc_sub.add_argument(
        '--threads',
        type=int,
        required=False,
        help='''
        Specify number of threads to use while executing models. Overrides
        settings in profiles.yml.
        '''
    )
    rpc_sub.add_argument(
        '--no-version-check',
        dest='sub_version_check',  # main cli arg precedes subcommands
        action='store_false',
        default=None,
        help='''
        If set, skip ensuring dbt's version matches the one specified in
        the dbt_project.yml file ('require-dbt-version')
        '''
    )

    if len(args) == 0:
        p.print_help()
        sys.exit(1)

    parsed = p.parse_args(args)
    # profiles_dir is set before subcommands and after, so normalize
    if hasattr(parsed, 'sub_profiles_dir'):
        if parsed.sub_profiles_dir is not None:
            parsed.profiles_dir = parsed.sub_profiles_dir
        delattr(parsed, 'sub_profiles_dir')
    if hasattr(parsed, 'profiles_dir'):
        if parsed.profiles_dir is None:
            parsed.profiles_dir = flags.PROFILES_DIR
        else:
            parsed.profiles_dir = os.path.abspath(parsed.profiles_dir)
            # needs to be set before the other flags, because it's needed to
            # read the profile that contains them
            flags.PROFILES_DIR = parsed.profiles_dir

    # version_check is set before subcommands and after, so normalize
    if hasattr(parsed, 'sub_version_check'):
        if parsed.sub_version_check is False:
            parsed.version_check = False
        delattr(parsed, 'sub_version_check')

    # fail_fast is set before subcommands and after, so normalize
    if hasattr(parsed, 'sub_fail_fast'):
        if parsed.sub_fail_fast is True:
            parsed.fail_fast = True
        delattr(parsed, 'sub_fail_fast')

    if getattr(parsed, 'project_dir', None) is not None:
        expanded_user = os.path.expanduser(parsed.project_dir)
        parsed.project_dir = os.path.abspath(expanded_user)

    if not hasattr(parsed, 'which'):
        # the user did not provide a valid subcommand. trigger the help message
        # and exit with a error
        p.print_help()
        p.exit(1)

    return parsed



import os
from dataclasses import dataclass
from typing import Iterable, Optional

from dbt.contracts.graph.manifest import SourceFile
from dbt.contracts.graph.nodes import RPCNode, Macro
from dbt.contracts.graph.unparsed import UnparsedMacro
from dbt.exceptions import InternalException
from dbt.node_types import NodeType
from dbt.parser.base import SimpleSQLParser
from dbt.parser.macros import MacroParser
from dbt.parser.search import FileBlock


@dataclass
class RPCBlock(FileBlock):
    rpc_name: str

    @property
    def name(self):
        return self.rpc_name


class RPCCallParser(SimpleSQLParser[RPCNode]):
    def parse_from_dict(self, dct, validate=True) -> RPCNode:
        if validate:
            RPCNode.validate(dct)
        return RPCNode.from_dict(dct)

    @property
    def resource_type(self) -> NodeType:
        return NodeType.RPCCall

    def get_compiled_path(cls, block: FileBlock):
        # we do it this way to make mypy happy
        if not isinstance(block, RPCBlock):
            raise InternalException(
                'While parsing RPC calls, got an actual file block instead of '
                'an RPC block: {}'.format(block)
            )

        return os.path.join('rpc', block.name)

    def parse_remote(self, sql: str, name: str, language: Optional[str] = 'sql') -> RPCNode:
        source_file = SourceFile.remote(sql, self.project.project_name, language)
        contents = RPCBlock(rpc_name=name, file=source_file)
        return self.parse_node(contents)


class RPCMacroParser(MacroParser):
    def parse_remote(self, contents) -> Iterable[Macro]:
        base = UnparsedMacro(
            path='from remote system',
            original_file_path='from remote system',
            package_name=self.project.project_name,
            raw_code=contents,
            language='sql',
            resource_type=NodeType.Macro,
        )
        for node in self.parse_unparsed_macros(base):
            yield node



import multiprocessing
from datetime import datetime
from typing import Optional, Union, MutableMapping
from typing_extensions import Protocol

import dbt.exceptions
from dbt_rpc.contracts.rpc import (
    TaskHandlerState,
    TaskID,
    TaskTags,
    TaskTiming,
    TaskRow,
)


class TaskHandlerProtocol(Protocol):
    task_id: TaskID
    state: TaskHandlerState
    started: Optional[datetime] = None
    ended: Optional[datetime] = None
    process: Optional[multiprocessing.Process] = None

    @property
    def request_id(self) -> Union[str, int]:
        pass

    @property
    def request_source(self) -> str:
        pass

    @property
    def timeout(self) -> Optional[float]:
        pass

    @property
    def method(self) -> str:
        pass

    @property
    def tags(self) -> Optional[TaskTags]:
        pass

    def _assert_started(self) -> datetime:
        if self.started is None:
            raise dbt.exceptions.InternalException(
                'task handler started but start time is not set'
            )
        return self.started

    def _assert_ended(self) -> datetime:
        if self.ended is None:
            raise dbt.exceptions.InternalException(
                'task handler finished but end time is not set'
            )
        return self.ended

    def make_task_timing(
        self, now_time: datetime
    ) -> TaskTiming:
        # get information about the task in a way that should not provide any
        # conflicting information. Calculate elapsed time based on `now_time`
        state = self.state
        # store end/start so 'ps' output always makes sense:
        # not started -> no start time/elapsed, running -> no end time, etc
        end = None
        start = None
        elapsed = None
        if state > TaskHandlerState.NotStarted:
            start = self._assert_started()
            elapsed_end = now_time

            if state.finished:
                elapsed_end = self._assert_ended()
                end = elapsed_end

            elapsed = (elapsed_end - start).total_seconds()
        return TaskTiming(state=state, start=start, end=end, elapsed=elapsed)

    def make_task_row(self, now_time: datetime) -> TaskRow:
        timing = self.make_task_timing(now_time)

        return TaskRow(
            task_id=self.task_id,
            request_id=self.request_id,
            request_source=self.request_source,
            method=self.method,
            state=timing.state,
            start=timing.start,
            end=timing.end,
            elapsed=timing.elapsed,
            timeout=self.timeout,
            tags=self.tags,
        )


TaskHandlerMap = MutableMapping[TaskID, TaskHandlerProtocol]



import os
import signal
from datetime import datetime
from typing import Type, Union, Any, List, Dict

import dbt.exceptions
from dbt_rpc.contracts.rpc import (
    TaskTags,
    StatusParameters,
    LastParse,
    GCParameters,
    GCResult,
    GetManifestResult,
    KillParameters,
    KillResult,
    KillResultStatus,
    PSParameters,
    TaskRow,
    PSResult,
    RemoteExecutionResult,
    RemoteFreshnessResult,
    RemoteRunResult,
    RemoteCompileResult,
    RemoteCatalogResults,
    RemoteDepsResult,
    RemoteRunOperationResult,
    PollParameters,
    PollResult,
    PollInProgressResult,
    PollKilledResult,
    PollExecuteCompleteResult,
    PollGetManifestResult,
    PollRunCompleteResult,
    PollCompileCompleteResult,
    PollCatalogCompleteResult,
    PollFreshnessResult,
    PollRemoteEmptyCompleteResult,
    PollRunOperationCompleteResult,
    TaskHandlerState,
    TaskTiming,
)
from dbt.logger import LogMessage
from dbt_rpc.rpc.error import dbt_error, RPCException
from dbt_rpc.rpc.method import RemoteBuiltinMethod
from dbt_rpc.rpc.task_handler import RequestTaskHandler


class GC(RemoteBuiltinMethod[GCParameters, GCResult]):
    METHOD_NAME = 'gc'

    def set_args(self, params: GCParameters):
        super().set_args(params)

    def handle_request(self) -> GCResult:
        if self.params is None:
            raise dbt.exceptions.InternalException('GC: params not set')
        return self.task_manager.gc_safe(
            task_ids=self.params.task_ids,
            before=self.params.before,
            settings=self.params.settings,
        )


class Kill(RemoteBuiltinMethod[KillParameters, KillResult]):
    METHOD_NAME = 'kill'

    def set_args(self, params: KillParameters):
        super().set_args(params)

    def handle_request(self) -> KillResult:
        if self.params is None:
            raise dbt.exceptions.InternalException('Kill: params not set')
        result = KillResult()
        task: RequestTaskHandler
        try:
            task = self.task_manager.get_request(self.params.task_id)
        except dbt.exceptions.UnknownAsyncIDException:
            # nothing to do!
            return result

        result.state = KillResultStatus.NotStarted

        if task.process is None:
            return result
        pid = task.process.pid
        if pid is None:
            return result

        if task.process.is_alive():
            result.state = KillResultStatus.Killed
            task.ended = datetime.utcnow()
            os.kill(pid, signal.SIGINT)
            task.state = TaskHandlerState.Killed
        else:
            result.state = KillResultStatus.Finished
            # the state must be "Completed"

        return result


class Status(RemoteBuiltinMethod[StatusParameters, LastParse]):
    METHOD_NAME = 'status'

    def set_args(self, params: StatusParameters):
        super().set_args(params)

    def handle_request(self) -> LastParse:
        return self.task_manager.last_parse


class PS(RemoteBuiltinMethod[PSParameters, PSResult]):
    METHOD_NAME = 'ps'

    def set_args(self, params: PSParameters):
        super().set_args(params)

    def keep(self, row: TaskRow):
        if self.params is None:
            raise dbt.exceptions.InternalException('PS: params not set')
        if row.state.finished and self.params.completed:
            return True
        elif not row.state.finished and self.params.active:
            return True
        else:
            return False

    def handle_request(self) -> PSResult:
        rows = [
            row for row in self.task_manager.task_table() if self.keep(row)
        ]
        rows.sort(key=lambda r: (r.state, r.start, r.method))
        result = PSResult(rows=rows, logs=[])
        return result


def poll_complete(
    timing: TaskTiming, result: Any, tags: TaskTags, logs: List[LogMessage]
) -> PollResult:
    if timing.state not in (TaskHandlerState.Success, TaskHandlerState.Failed):
        raise dbt.exceptions.InternalException(
            f'got invalid result state in poll_complete: {timing.state}'
        )

    cls: Type[Union[
        PollExecuteCompleteResult,
        PollRunCompleteResult,
        PollCompileCompleteResult,
        PollCatalogCompleteResult,
        PollRemoteEmptyCompleteResult,
        PollRunOperationCompleteResult,
        PollGetManifestResult,
        PollFreshnessResult,
    ]]

    if isinstance(result, RemoteExecutionResult):
        cls = PollExecuteCompleteResult
    # order matters here, as RemoteRunResult subclasses RemoteCompileResult
    elif isinstance(result, RemoteRunResult):
        cls = PollRunCompleteResult
    elif isinstance(result, RemoteCompileResult):
        cls = PollCompileCompleteResult
    elif isinstance(result, RemoteCatalogResults):
        cls = PollCatalogCompleteResult
    elif isinstance(result, RemoteDepsResult):
        cls = PollRemoteEmptyCompleteResult
    elif isinstance(result, RemoteRunOperationResult):
        cls = PollRunOperationCompleteResult
    elif isinstance(result, GetManifestResult):
        cls = PollGetManifestResult
    elif isinstance(result, RemoteFreshnessResult):
        cls = PollFreshnessResult
    else:
        raise dbt.exceptions.InternalException(
            'got invalid result in poll_complete: {}'.format(result)
        )
    return cls.from_result(result, tags, timing, logs)


def _dict_logs(logs: List[LogMessage]) -> List[Dict[str, Any]]:
    return [log.to_dict(omit_none=True) for log in logs]


class Poll(RemoteBuiltinMethod[PollParameters, PollResult]):
    METHOD_NAME = 'poll'

    def set_args(self, params: PollParameters):
        super().set_args(params)

    def handle_request(self) -> PollResult:
        if self.params is None:
            raise dbt.exceptions.InternalException('Poll: params not set')
        task_id = self.params.request_token
        task: RequestTaskHandler = self.task_manager.get_request(task_id)

        task_logs: List[LogMessage] = []
        if self.params.logs:
            task_logs = task.logs[self.params.logs_start:]

        # Get a state and store it locally so we ignore updates to state,
        # otherwise things will get confusing. States should always be
        # "forward-compatible" so if the state has transitioned to error/result
        # but we aren't there yet, the logs will still be valid.

        timing = task.make_task_timing(datetime.utcnow())
        state = timing.state
        if state <= TaskHandlerState.Running:
            return PollInProgressResult(
                tags=task.tags,
                logs=task_logs,
                state=timing.state,
                start=timing.start,
                end=timing.end,
                elapsed=timing.elapsed,
            )
        elif state == TaskHandlerState.Error:
            err = task.error
            if err is None:
                exc = dbt.exceptions.InternalException(
                    f'At end of task {task_id}, error state but error is None'
                )
                raise RPCException.from_error(
                    dbt_error(exc, logs=_dict_logs(task_logs))
                )
            # the exception has logs already attached from the child, don't
            # overwrite those
            raise err
        elif state in (TaskHandlerState.Success, TaskHandlerState.Failed):

            if task.result is None:
                exc = dbt.exceptions.InternalException(
                    f'At end of task {task_id}, state={state} but result is '
                    'None'
                )
                raise RPCException.from_error(
                    dbt_error(exc, logs=_dict_logs(task_logs))
                )
            return poll_complete(
                timing=timing,
                result=task.result,
                tags=task.tags,
                logs=task_logs
            )
        elif state == TaskHandlerState.Killed:
            return PollKilledResult(
                tags=task.tags,
                logs=task_logs,
                state=timing.state,
                start=timing.start,
                end=timing.end,
                elapsed=timing.elapsed,
            )
        else:
            exc = dbt.exceptions.InternalException(
                f'Got unknown value state={state} for task {task_id}'
            )
            raise RPCException.from_error(
                dbt_error(exc, logs=_dict_logs(task_logs))
            )
